## Создание клиента

Вся работа с API НСПД происходит через объект клиента. Работа с ними возможна как в контекстном менеджере:
```python
from pynspd import Nspd

with Nspd() as nspd:
    # Ваш код здесь...
```

Либо же мы можем создать объект клиента без контекстного менеджера, но не забудьте потом завершить его сессию:

```python
from pynspd import Nspd

nspd = Nspd()

# Ваш код здесь...

nspd.close()
```

О том, почему важно закрывать сессию, вы можете прочитать в документации [httpx](https://www.python-httpx.org/advanced/clients/#usage) - замечательной библиотеке для http-запросов, которую `pynspd` использует под капотом.

## Поиск объектов

### По поисковому запросу

НСПД предоставляет два вида поиска объектов - по виду объекта и по слою.

#### По виду объекта

```python
from pynspd import ThemeId

features = nspd.search("77:05:0001005:19", ThemeId.REAL_ESTATE_OBJECTS)
print(features[0].properties.options.land_record_type)
#> Земельный участок
```

#### По слоям

```python
features = nspd.search_in_layer(
    "77:05:0001005:19", 
    NspdFeature.by_title("Земельные участки из ЕГРН")
)
```

Если вы работаете в современной **IDE** (*VS Code*, *PyCharm* и т.д.), то она сама подскажет вам доступный список слоев:

![Layer titles example](assets/layer_assist.png)

Кроме того этот метод приводит ответ от НСПД к статическому типу и мы можем проверить доступные поля: 

![Typed answer example](assets/typed_answer.png)

Подробнее о типизации ответов читайте в [этой статье](advanced/typing.md).


#### Точное совпадение

Довольно часто требуется найти информацию по конкретному кадастровому номеру. В то же время методы `.search(...)`
и `.search_in_layer(...)` выдают массив (или `None`, если ничего не найдено), что порождает лишний код
по фильтрации массива и извлечения из него.

Чтобы сократить вышеописанные кейсы, вы можете использовать методы `.find(...)` и `.find_in_layer(...)`, 
которые всегда возвращают либо один объект, либо `None`:

```python
feat = nspd.find("77:05:0001005:19")
print(feat.properties.options.land_record_type)
#> Земельный участок
```

??? info "Немного подробнее про фильтрацию в `.find`"
    `.find` не просто проверяет, что в массиве один элемент и возвращает его.
    Он проверяет, что поисковой запрос действительно присутствует в ответе и 
    НСПД не галлюцинирует (например, запрос `77:1:3033:1031` возвращает данные по к/н `02:54:000000:81`),
    а так же напрямую относится к объекту (например, исключает помещения, включенные в ответ по зданию)

### В точке

```python
from shapely import Point

layer_def = NspdFeature.by_title("Земельные участки из ЕГРН")
feats = nspd.search_at_point(Point(37.546440653, 55.787139958), layer_def)

print(feats[0].properties.options.cad_num)
#> "77:09:0005008:11446"
```

!!! tip
    Если вы не хотите использовать `shapely.Point`, вы можете использовать метод `.search_at_coords(lat, lng, layer_def)`.
    Обратите внимание, что в географических координатах сначала идет широта, т.е. `lat`=`y`

Аналогично поиску по слою, ответ будет приведен к статическому типу.

### В контуре

```python
from shapely import from_wkt

contour = from_wkt(
    "Polygon ((37.62381 55.75345, 37.62577 55.75390, 37.62448 55.75278, 37.62381 55.75345))"
)

feats = nspd.search_in_contour(
    contour,
    NspdFeature.by_title("Земельные участки из ЕГРН"),
)
cns = [i.properties.options.cad_num for i in feats]
print(cns)
#> ["77:01:0001011:8", "77:01:0001011:14", "77:01:0001011:16"]
```

Аналогично поиску по слою, ответ будет приведен к статическому типу.

Данный метод передает ответ НСПД "как есть". Это значит, что если площадь поиска слишком большая, 
будет вызвана ошибка `TooBigContour`. Однако, вы можете воспользоваться генератором `.search_in_contour_iter(...)`,
который обходит известные ограничения:

```python
for i in nspd.search_in_contour_iter(
    contour,
    NspdFeature.by_title("Земельные участки из ЕГРН"),
    only_intersects=True, # по умолчанию метод ищет все объекты 
                          # в bounding box контура для экономии ресурсов
):
    print(i.properties.options.cad_num)
```

## Работа с геометрией

Исходный формат ответов по объекту от НСПД - [GeoJSON](https://ru.wikipedia.org/wiki/GeoJSON). 

!!! info "Системы коодинат"
    Несмотря на то, что НСПД может в разных ситуациях ответить в разных системах координат, `pynspd` ВСЕГДА приводит систему коодинат геометрии в EPSG:4326.

    По этой же причине все методы, работающие с геометрией, ожидают, что она будет в СК EPSG:4326.

Однако, если нам необходимо производить геометрические операции, он не совсем подходит. Поэтому `pynspd` предоставляет возможность конвертировать геометрию в формат [shapely](https://shapely.readthedocs.io/en/2.0.4/manual.html):

```python
feat.geometry.to_shape()
```

Или же сразу в мультигеометрию (что может быть удобно для сохранения в базу данных, вроде [PostGIS](https://postgis.net/)):

```python
feat.geometry.to_multi_shape()
```

!!! warning "Участки без координат границ"
    Попытка приведения объекта без координат границ к мультитипу приведет к ошибке. Вы можете заранее проверить наличие координат по свойству `feat.properties.options.no_coords`


## Работа со свойствами

Свойства объекта лежат в свойстве `.properties`:

```python
feat = nspd.search_in_layer(
    '63:01:0810003:510', 
    NspdFeature.by_title("Земельные участки из ЕГРН")
)

print(feat.properties.model_dump().keys())
#> ['category',
#>  'category_name',
#>  'options',
#>  'system_info',
#>  ...]
```

!!! info "Что за `.model_dump()`?"
    Все результаты `pynspd` является [pydantic](https://docs.pydantic.dev/latest)-моделями. 
    Фактически это значит, что `pynspd` отдает вам [датакласс](https://docs.python.org/3/library/dataclasses.html), 
    прошедший предварительную валидацию. 
    Для конвертации `pydantic`-модели в словарь используется метод `.model_dump()`

Как можно заметить, здесь нет информации о самом объекте - дело в том, что НСПД передает ее во вложенном поле `.options`

```python
print(feat.properties.options.model_dump().keys())
#> ['cad_num',
#>  'readable_address',
#>  'specified_area',
#>  'declared_area',
#>  'cost_value',
#>  ...]
```

Также вы можете сразу получить словарь с названиями из карточки НСПД:

```python
print(feat.properties.cast().options.model_dump_human_readable().keys())
#> ['Кадастровый номер',
#>  'Адрес',
#>  'Площадь уточненная (кв. м)',
#>  'Площадь декларированная (кв. м)',
#>  'Площадь (кв. м)',
#>  'Кадастровая стоимость (руб.)',
#>  ...]
```

!!! info "Приведение типа объекта"
    Если вас заинтересовало, почему используется метод `.cast()`, 
    вы можете узнать об этом в статье ["Типизация ответов"](advanced/typing.md)

## Получении информации из вкладок

Часть информации об объекте доступна только в отдельных вкладках:

![NSPD tabs example](assets/nspd_tabs.png)

Для получения этой информации при помощи `pynspd`, нужно:

1. Получить объект любым вышеописанным способом;
2. Вызвать нижеперечисленный метод, передав в него объект.

Перечень существующих методов API для получения данных из вкладок:

- `nspd.tab_land_parts(...)` - Части ЗУ
- `nspd.tab_land_links(...)` - Связанные ЗУ
- `nspd.tab_permission_type(...)` - Виды разрешенного использования
- `nspd.tab_composition_land(...)` - Состав ЕЗП
- `nspd.tab_build_parts(...)` - Части ОКС
- `nspd.tab_objects_list(...)` - Объекты